"""Time utilities for ionospheric analysis."""

from __future__ import annotations

import astropy.units as u
import numpy as np
from astropy.time import Time
from numpy.typing import NDArray


def get_gps_week(time: Time) -> NDArray[np.int32]:
    """Get the GPS week from a time.

    Parameters
    ----------
    time : Time
        Time(s) to get the GPS week from

    Returns
    -------
    NDArray
        GPS week(s)
    """
    return np.array(np.floor((time.gps * u.s).to(u.week).value), dtype=np.int32)


def get_unique_days(times: Time) -> Time:
    """Get the unique days from a list of times.

    Parameters
    ----------
    times : Time
        Times to get the unique days from.

    Returns
    -------
    Time
        Unique days
    """
    return Time(np.sort(np.unique(np.floor(times.mjd))), format="mjd")


def get_unique_days_index(times: Time) -> NDArray[np.int32]:
    """Get the unique days index from a list of times.

    Parameters
    ----------
    times : Time
        Times to get the unique days index from.

    Returns
    -------
    NDArray
        Unique days index
    """
    return np.array(
        np.searchsorted(get_unique_days(times).mjd, np.floor(times.mjd)), dtype=np.int32
    )


# TODO: Add docs
def get_indexlist_unique_days(unique_days: Time, times: Time) -> NDArray[np.bool_]:
    return np.array(
        np.floor(times.mjd)[np.newaxis] == unique_days.mjd[:, np.newaxis],
        dtype=np.bool_,
    )


# TODO: Add docs
def get_consecutive_days(unique_days: Time) -> NDArray[np.int32]:
    return np.cumsum(np.diff(unique_days.mjd, prepend=unique_days.mjd[0]) > 1)



"""Some useful multipurpose functions to interpolate on grids"""

from __future__ import annotations

from typing import NamedTuple

import numpy as np
from numpy.typing import NDArray


class Indices(NamedTuple):
    """Indices of the closest two points in a possibly wrapping selection and the inverse distance weights"""

    idx1: NDArray[np.int64]
    """Index of the first closest point"""
    idx2: NDArray[np.int64]
    """Index of the second closest point"""
    w1: NDArray[np.float64]
    """Weight of the first closest point"""
    w2: NDArray[np.float64]
    """Weight of the second closest point"""


class SortedIndices(NamedTuple):
    """Indices of the closest two points in a possibly wrapping selection"""

    indices: NDArray[np.int64]
    """Index of the first closest point"""
    distance: NDArray[np.float64]
    """Index of the second closest point"""


class Weights(NamedTuple):
    """Weights of the closest two points in a possibly wrapping selection"""

    w1: NDArray[np.float64]
    """Weight of the first closest point"""
    w2: NDArray[np.float64]
    """Weight of the second closest point"""


# TODO: Get typing right for this function
def get_indices_axis(
    goal: NDArray[np.float64], selection: NDArray[np.float64], wrap_unit: float = 0
) -> Indices:
    """get indices of the closest two points in a possibly wrapping selection for an array
    of goals"""
    if wrap_unit > 0:
        idx1 = np.argmin(
            np.absolute(
                np.remainder(
                    goal[..., np.newaxis] - selection + 0.5 * wrap_unit, wrap_unit
                )
                - 0.5 * wrap_unit
            ),
            axis=-1,
        ).astype(np.int64)
    else:
        idx1 = np.argmin(np.absolute(goal[..., np.newaxis] - selection), axis=-1)
    if np.isscalar(idx1):
        if goal < selection[idx1]:
            idx2 = idx1
            idx1 -= 1
            if idx1 < 0:
                idx1 = selection.shape[0] if wrap_unit > 0 else 0
        else:
            idx2 = idx1 + 1
            if idx2 >= selection.shape[0]:
                if wrap_unit > 0:
                    idx2 = 0
                else:
                    idx2 -= 1
    else:
        idx2 = np.copy(idx1)
        idx1[goal < selection[idx1]] = idx1[goal < selection[idx1]] - 1
        idx2[goal >= selection[idx2]] = idx2[goal >= selection[idx2]] + 1
        if wrap_unit > 0:
            idx1[idx1 < 0] = selection.shape[0] - 1
            idx2[idx2 >= selection.shape[0]] = 0
        else:
            idx1[idx1 < 0] = 0
            idx2[idx2 >= selection.shape[0]] -= 1
    weights = _get_weights(goal, idx1, idx2, selection, wrap_unit)
    return Indices(idx1=idx1, idx2=idx2, w1=weights.w1, w2=weights.w2)


# TODO: Get typing right for this function
def _get_weights(
    goal: float,
    index1: NDArray[np.int64],
    index2: NDArray[np.int64],
    selection: NDArray[np.float64],
    wrap_unit: float = 0,
) -> Weights:
    """Calculate weights based on distance of goal to selection

    Parameters
    ----------
    goal : NDArray
        array of points to get weights for
    index1 : NDArray
        indices in selection for goals (index1, index2) per goal
    index2 : NDArray
        indices in selection for goals (index1, index2) per goal
    selection : NDArray
        array to select from
    wrap_unit : float, optional
        if goal/selection is a wrapable (e.g. angle) set this unit (e.g. 360), by default 0

    Returns
    -------
    namedtuple
        tuple with weights
    """
    distance1 = np.abs(wrap_around_zero(selection[index1] - goal, wrap_unit))
    distance2 = np.abs(wrap_around_zero(selection[index2] - goal, wrap_unit))
    sumdist = distance1 + distance2
    return Weights(w1=1 - distance1 / sumdist, w2=1 - distance2 / sumdist)


def get_indices(
    goal: float, selection: NDArray[np.float64], wrap_unit: float = 0
) -> Indices:
    """find the indices of the closest two points in a possibly wrapping array selection

    Parameters
    ----------
    goal : float
        location of point
    selection : NDArray
        array of points
    wrap_unit : float, optional
        if goal/selection is a wrapping entity (e.g. angles) set this to the wrap value (e.g. 360), by default 0

    Returns
    -------
    Indices:
        sorted list of index1 and index2
    """

    if wrap_unit > 0:
        idx1 = np.argmin(
            np.absolute(
                np.remainder(goal - selection + 0.5 * wrap_unit, wrap_unit)
                - 0.5 * wrap_unit
            )
        )
    else:
        idx1 = np.argmin(
            np.absolute(goal - selection),
        )
    idx2 = idx1 - 1 if goal < selection[idx1] else idx1 + 1
    if idx2 < 0 or idx2 >= selection.shape[0]:
        idx2 = idx1
    weights = _get_weights(goal, idx1, idx2, selection, wrap_unit)
    return Indices(idx1=idx1, idx2=idx2, w1=weights.w1, w2=weights.w2)


def get_sorted_indices(
    lon: float,
    lat: float,
    avail_lon: NDArray[np.float64],
    avail_lat: NDArray[np.float64],
    wrap_unit: float = 360.0,
) -> SortedIndices:
    """find distances of a lon/lat grid to a point and return sorted list of indices"""
    distance = (
        wrap_around_zero(avail_lon - lon, wrap_unit) ** 2
        + wrap_around_zero(avail_lat - lat, wrap_unit) ** 2
    )
    sorted_idx = np.argsort(distance)
    return SortedIndices(indices=sorted_idx, distance=distance[sorted_idx])


def get_interpol(data: NDArray[np.float64], dist: NDArray[np.float64]) -> float:
    """get distance weighted sum of data"""
    w = 1.0 / dist
    w /= np.sum(w)
    return float(np.sum(data * w))


def wrap_around_zero(
    data: NDArray[np.float64], wrap_unit: float = 2 * np.pi
) -> NDArray[np.float64]:
    """Function to calculate the remainder of data such that this is centered around zero"""
    return np.remainder(data + 0.5 * wrap_unit, wrap_unit) - 0.5 * wrap_unit


def compute_index_and_weights(
    maparray: NDArray[np.float64], mapvalues: NDArray[np.float64]
) -> Indices:
    """helper function  to get indices and weights for interpolating tecmaps


    Args:
        maparray (NDArray) : array to get indices in
        mapvalues (Union[float,np.array]) :  values to get indices for
    Returns:
        Tuple[np.array, np.array, np.array]: idx1,idx2 and weights for idx2,
                                             idx2 is always >= idx1

    """
    is_reverse = maparray[1] < maparray[0]
    idx1 = np.argmin(
        np.absolute(maparray[np.newaxis] - mapvalues[:, np.newaxis]), axis=1
    )
    idx2 = idx1.copy()
    if not is_reverse:
        idx1[maparray[idx1] > mapvalues] -= 1
        idx2[maparray[idx2] < mapvalues] += 1
    else:
        idx1[maparray[idx1] < mapvalues] -= 1
        idx2[maparray[idx2] > mapvalues] += 1
    idx1[idx1 < 0] = 0
    idx2[idx2 < 0] = 0
    idx1[idx1 >= maparray.shape[0]] = maparray.shape[0] - 1
    idx2[idx2 >= maparray.shape[0]] = maparray.shape[0] - 1
    _steps = np.absolute(maparray[idx2] - maparray[idx1])
    weights = np.absolute(mapvalues - maparray[idx1])
    weights[_steps == 0] = 1.0
    weights[_steps != 0] = weights[_steps != 0] / _steps[_steps != 0]
    return Indices(idx1=idx1, idx2=idx2, w1=1 - weights, w2=weights)



#!/usr/bin/env python
from __future__ import annotations

import argparse
from pathlib import Path

from spinifex.vis_tools.ms_tools import (
    cli_get_dtec_h5parm_from_ms,
    cli_get_rm_h5parm_from_ms,
)


def get_parser():
    parser = argparse.ArgumentParser(
        description="This Spinifex command-line interface can make an H5Parm with ionospheric data (TEC or RM), using the metadata from a measurement set."
    )
    subparsers = parser.add_subparsers(dest="command")

    # Subparser for get_rm_h5parm_from_ms
    parser_rm = subparsers.add_parser(
        "get_rm_h5parm_from_ms",
        description="Calculate RM values using spinifex, write to h5parm",
    )
    parser_rm.add_argument(
        "ms",
        type=Path,
        help="Measurement set for which the RM values should be calculated.",
    )
    parser_rm.add_argument(
        "--solset-name",
        type=str,
        help="Solset name. Default: create a new one based on first existing sol###",
    )
    parser_rm.add_argument(
        "--soltab-name", type=str, help="Soltab name. Default: rotationmeasure"
    )
    parser_rm.add_argument(
        "-o",
        "--h5parm",
        default="rotationmeasure.h5",
        type=Path,
        help="h5parm to which the results of the RotationMeasure is added.",
    )
    parser_rm.add_argument(
        "-a",
        "--add-to-existing-solset",
        action="store_true",
        help="Add to existing solset if it exists",
    )
    parser_rm.set_defaults(func=cli_get_rm_h5parm_from_ms)

    # Subparser for get_tec_h5parm_from_ms
    parser_tec = subparsers.add_parser(
        "get_tec_h5parm_from_ms",
        description="Calculate tec values using spinifex, write to h5parm",
    )
    parser_tec.add_argument(
        "ms",
        type=Path,
        help="Measurement set for which the tec values should be calculated.",
    )
    parser_tec.add_argument(
        "--solset-name",
        type=str,
        help="Solset name. Default: create a new one based on first existing sol###",
    )
    parser_tec.add_argument(
        "--soltab-name", type=str, help="Soltab name. Default: tec###"
    )
    parser_tec.add_argument(
        "-o",
        "--h5parm",
        default="tec.h5",
        type=Path,
        help="h5parm to which the results of the tec is added.",
    )
    parser_tec.add_argument(
        "-a",
        "--add-to-existing-solset",
        action="store_true",
        help="Add to existing solset if it exists",
    )
    parser_tec.set_defaults(func=cli_get_dtec_h5parm_from_ms)

    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    if args.command:
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



"""Logging utilities for spinifex."""

from __future__ import annotations

import io
import logging

logging.captureWarnings(True)


formatter = logging.Formatter(
    fmt="%(levelname)s %(module)s - %(funcName)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class TqdmToLogger(io.StringIO):
    """
    Output stream for TQDM which will output to logger module instead of
    the StdOut.
    """

    logger = None
    level = None
    buf = ""

    def __init__(self, logger: logging.Logger | None, level: int | None = None) -> None:
        super().__init__()
        self.logger = logger
        self.level = level or logging.INFO

    def write(self, buf: str) -> int:
        self.buf = buf.strip("\r\n\t ")
        return len(buf)

    def flush(self) -> None:
        if self.logger is not None and isinstance(self.level, int):
            self.logger.log(self.level, self.buf)


def set_verbosity(verbosity: int) -> None:
    """Set the logger verbosity.

    Parameters
    ----------
    verbosity : int
        Verbosity level
    """
    if verbosity == 0:
        level = logging.WARNING
    elif verbosity == 1:
        level = logging.INFO
    elif verbosity >= 2:
        level = logging.DEBUG
    else:
        level = logging.CRITICAL

    logging.getLogger().setLevel(level)
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    ch.setLevel(level)
    logging.getLogger().addHandler(ch)


logger = logging.getLogger("spinifex")

ch = logging.StreamHandler()
ch.setFormatter(formatter)
ch.setLevel(logging.INFO)
logger.addHandler(ch)
logger.setLevel(logging.INFO)



"""Module for getting the Ionospheric Piercepoints"""

from __future__ import annotations

from typing import NamedTuple

import astropy.units as u
import numpy as np
from astropy.constants import R_earth
from astropy.coordinates import ITRS, AltAz, EarthLocation, SkyCoord
from astropy.time import Time
from numpy.typing import NDArray

from spinifex.logger import logger


class IPP(NamedTuple):
    """Ionospheric Piercepoints"""

    loc: EarthLocation
    """location of the piercepoints, dimension: times x altitudes. All altitudes are assumed to be equal"""
    times: Time
    """array of times"""
    los: SkyCoord
    """Line of sight direction in ITRS coordinates"""
    airmass: NDArray[np.float64]
    """airmass factor to convert to slant values"""
    altaz: AltAz
    """azimuth elevation"""
    station_loc: EarthLocation
    """Observer Location"""


def get_ipp_from_skycoord(
    loc: EarthLocation, times: Time, source: SkyCoord, height_array: u.Quantity
) -> IPP:
    """Get ionospheric piercepoints

    Parameters
    ----------
    loc : EarthLocation
        observer location
    times : Time
        observation times
    source : SkyCoord
        source location
    height_array : u.Quantity
        array of altitudes

    Returns
    -------
    IPP
        Ionospheric piercepoints
    """
    # Note: at the moment we calculate ipp per station. I think this is ok,
    # but maybe we need to include a many stations option
    aa = AltAz(location=loc, obstime=times)
    altaz = source.transform_to(aa)
    return get_ipp_from_altaz(loc, altaz, height_array)


def get_ipp_from_altaz(
    loc: EarthLocation, altaz: AltAz, height_array: u.Quantity
) -> IPP:
    """get ionospheric piercepoints from azimuth elevations

    Parameters
    ----------
    loc : EarthLocation
        observer location
    altaz : AltAz
        azimuth and elevations for all times
    height_array : u.Quantity
        array of altitudes

    Returns
    -------
    IPP
        ionospheric piercepoints
    """
    if not altaz.obstime.shape or altaz.obstime.shape != altaz.az.shape:
        altaz = _make_dimensions_match(altaz)
    los_dir = altaz.transform_to(ITRS(obstime=altaz.obstime))
    ipp, airmass = _get_ipp_simple(height_array=height_array, loc=loc, los_dir=los_dir)
    return IPP(
        loc=EarthLocation.from_geocentric(*ipp),
        times=altaz.obstime,
        los=los_dir,
        airmass=airmass,
        altaz=altaz,
        station_loc=loc,
    )


def _make_dimensions_match(altaz: AltAz) -> AltAz:
    """Helper function to change time dimensions suchthat they correspond to the altaz dimension

    Parameters
    ----------
    altaz : AltAz
        the altaz object

    Returns
    -------
    AltAz
        altaz object with matching obstime dimension

    Raises
    ------
    NotImplementedError
        multiple times with different shape than altaz is not implemented yet
    """
    times = altaz.obstime
    az = altaz.az
    # if multiple azimuth/altitudes for one time, just increase dimensions of time
    if not times.shape:
        times = Time(times.mjd * np.ones(az.shape), format="mjd")
    if times.shape != az.shape:
        msg = (
            "Support for multiple times for azimuth/elevation grids is not implemented"
        )
        raise NotImplementedError(msg)

    return AltAz(az=altaz.az, alt=altaz.alt, obstime=times, location=altaz.location)


# TODO: Create return type for this function
def _get_ipp_simple(
    height_array: u.Quantity, loc: EarthLocation, los_dir: SkyCoord
) -> tuple[list[u.Quantity], NDArray[np.float64]]:
    r"""helper function to calculate ionospheric piercepoints using a simple spherical earth model

    .. code-block::

        |loc + alphas * los_dir| = R_earth + height_array

    solve for alphas using abc formula

    Parameters
    ----------
    height_array : u.Quantity
        array of altitudes
    loc : EarthLocation
        observer location
    los_dir : ITRS
        line of sight, unit vector

    Returns
    -------
    tuple(list[u.Quantity], NDArray)
        ipp.x, ipp.y, ipp.z positions, airmass
    """
    logger.info("Calculating ionospheric piercepoints")
    c_value = R_earth**2 - (R_earth + height_array) ** 2
    b_value = u.Quantity(loc.geocentric) @ los_dir.cartesian.xyz
    b_value = b_value[:, np.newaxis]
    alphas = -b_value + np.sqrt(b_value**2 - c_value)
    ipp = (
        u.Quantity(loc.geocentric)[:, np.newaxis, np.newaxis]
        + alphas * los_dir.cartesian.xyz[:, :, np.newaxis]
    )
    inv_airmass = np.einsum("ijk,ij->jk", ipp, los_dir.cartesian.xyz)
    inv_airmass /= R_earth + height_array  # normalized
    airmass = (
        1.0 / inv_airmass.decompose().value
    )  # if you forget the .decompose it can have airmass in (m/km)
    return ipp, airmass



from __future__ import annotations


class SpinifexError(Exception):
    """Base class for errors in spinifex"""


class IonexError(SpinifexError):
    """Error in IONEX files."""


class TimeResolutionError(IonexError):
    """Error in IONEX resolution."""


class FITSHeaderError(SpinifexError):
    """Error in FITS header"""



"""Several implementations of Ionospheric Models. They all should have the get_density function"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, Protocol, TypeVar

import astropy.units as u
import numpy as np
from astropy.coordinates import EarthLocation
from numpy.typing import NDArray
from pydantic import ValidationError

import spinifex.ionospheric.iri_density as iri
from spinifex.geometry.get_ipp import IPP
from spinifex.ionospheric.ionex_download import IonexOptions
from spinifex.ionospheric.ionex_manipulation import get_density_ionex
from spinifex.logger import logger

# TODO: add more ionospheric models
IonoOptions = IonexOptions  # Single model for now, but we can add more later e.g. IonexOptions | NewOptionType | AnotherOptionType
O = TypeVar("O", bound=IonoOptions)  # noqa: E741
O_contra = TypeVar("O_contra", contravariant=True, bound=IonoOptions)


class ModelDensityFunction(Protocol, Generic[O_contra]):
    """Model density callable"""

    def __call__(
        self,
        ipp: IPP,
        height: u.Quantity = 350 * u.km,
        options: O_contra | None = None,
    ) -> NDArray[np.float64]: ...


@dataclass
class IonosphericModels:
    """
    Names space for different ionospheric ionospheric_models. An ionospheric model should be
        a callable get_density
    """

    ionex: ModelDensityFunction[IonexOptions]
    ionex_iri: ModelDensityFunction[IonexOptions]


def get_density_ionex_single_layer(
    ipp: IPP, height: u.Quantity = 350 * u.km, options: IonexOptions | None = None
) -> NDArray[np.float64]:
    """gets the ionex files and interpolate values for a single altitude, thin screen assumption

    Parameters
    ----------
    ipp : IPP
        ionospheric piercepoints
    height : u.Quantity, optional
        altitude of the thin screen, by default 350*u.km
    ionex_options: IonexOptions | None, optional
        options for the ionospheric model, by default None

    Returns
    -------
    NDArray
        interpolated vTEC values at ipp, zeros everywhere apart from the altitude
        closest to the specified height
    """
    n_times = ipp.times.shape[0]  # we assume time is first axis
    index = np.argmin(
        np.abs(ipp.loc.height.to(u.km).value - height.to(u.km).value), axis=1
    )
    single_layer_loc = EarthLocation(ipp.loc[np.arange(n_times), index])
    ipp_single_layer = IPP(
        loc=single_layer_loc,
        times=ipp.times,
        los=ipp.los,
        airmass=ipp.airmass[:, index],
        altaz=ipp.altaz,
        station_loc=ipp.station_loc,
    )
    result = np.zeros(ipp.loc.shape, dtype=float)
    result[np.arange(n_times), index] = get_density_ionex(
        ipp_single_layer,
        ionex_options=options,
    )
    return result


def get_density_ionex_iri(
    ipp: IPP,
    height: u.Quantity = 350 * u.km,
    options: IonexOptions | None = None,
) -> NDArray[np.float64]:
    """gets the ionex files and interpolate values for a single altitude, then multiply with a
    normalised density profile from iri

    Parameters
    ----------
    ipp : IPP
        ionospheric piercepoints
    height : u.Quantity, optional
        altitude of the thin screen, by default 350*u.km
    ionex_options: IonexOptions | None, optional
        options for the ionospheric model, by default None

    Returns
    -------
    NDArray
        interpolated vTEC values at ipp
    """
    profile = iri.get_profile(ipp)
    tec = get_density_ionex_single_layer(ipp, height=height, options=options)
    # get tec at single altitude
    return np.array(np.sum(tec, keepdims=True, axis=1) * profile)


ionospheric_models = IonosphericModels(
    ionex=get_density_ionex_single_layer,
    ionex_iri=get_density_ionex_iri,
)


# TODO: Use the TypeVar `O` when we have more than one ionospheric model
# i.e.
# >>> def parse_iono_kwargs(
# >>>     iono_model: ModelDensityFunction[O], **kwargs: Any
# >>> ) -> O | None:
def parse_iono_kwargs(
    iono_model: ModelDensityFunction[IonoOptions], **kwargs: Any
) -> IonoOptions:
    """parse ionospheric options

    Parameters
    ----------
    iono_model : ModelDensityFunction
        ionospheric model
    **kwargs : Any
        options for the ionospheric model

    Raises
    ------
    TypeError
        Incorrect arguments for ionospheric model

    Returns
    -------
    IonoOptions
        ionospheric model options

    """

    try:
        if iono_model in (ionospheric_models.ionex, ionospheric_models.ionex_iri):
            options = IonexOptions(**kwargs)
            logger.info(f"Using ionospheric model {iono_model} with options {options}")
            return options

        # TODO: add more ionospheric models and their options
        msg = f"Unknown ionospheric model {iono_model}."
        raise TypeError(msg)

    except ValidationError as e:
        msg = f"Incorrect arguments {kwargs} for ionospheric model {iono_model}"
        raise TypeError(msg) from e


def parse_iono_model(iono_model_name: str) -> ModelDensityFunction[IonoOptions]:
    """parse ionospheric model name

    Parameters
    ----------
    iono_model_name : str
        name of the ionospheric model

    Returns
    -------
    ModelDensityFunction
        ionospheric model

    Raises
    ------
    TypeError
        if the ionospheric model is not known

    """

    try:
        return getattr(ionospheric_models, iono_model_name)  # type: ignore[no-any-return]
    except AttributeError as e:
        msg = f"Unknown ionospheric model {iono_model_name}. Supported models are {list(ionospheric_models.__annotations__.keys())}"
        raise TypeError(msg) from e



from __future__ import annotations

from astropy.utils import iers

iers.conf.auto_download = False

import astropy.units as u
import numpy as np
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from spinifex.geometry.get_ipp import IPP, get_ipp_from_skycoord


def get_test_ipp() -> IPP:
    cas_a = SkyCoord(ra=350.85 * u.deg, dec=58.815 * u.deg)
    lon = 6.367 * u.deg
    lat = 52.833 * u.deg
    heights = np.arange(100, 2000, 100) * u.km
    dwingeloo = EarthLocation(lon=lon, lat=lat, height=0 * u.km)
    times = Time("2020-01-20T01:00:00") + np.arange(0, 10) * 15 * u.min
    return get_ipp_from_skycoord(
        loc=dwingeloo, times=times, source=cas_a, height_array=heights
    )


def test_geometry():
    """Test the geometry module"""
    ipp = get_test_ipp()
    assert isinstance(ipp, IPP)
    assert isinstance(ipp.airmass, np.ndarray)



#  Copyright (C) 2024 ASTRON (Netherlands Institute for Radio Astronomy)
#  SPDX-License-Identifier: Apache-2.0

"""Module for getting ionospheric Faraday rotation from external geomagnetic and ionospheric ionospheric_models"""

from __future__ import annotations

from importlib import metadata

__version__ = metadata.version("spinifex")
from spinifex import options
from spinifex.ionospheric import ModelDensityFunction, ionospheric_models

__all__ = ["ModelDensityFunction", "ionospheric_models", "options"]



"""Utilities for downloading IONEX files from various servers."""

from __future__ import annotations

import asyncio
import shutil
from datetime import datetime
from enum import Enum
from ftplib import FTP
from pathlib import Path
from typing import Literal
from urllib.parse import urlparse

import astropy.units as u
import requests
from astropy.time import Time
from pydantic import Field

from spinifex.asyncio_wrapper import sync_wrapper
from spinifex.exceptions import IonexError, TimeResolutionError
from spinifex.logger import logger
from spinifex.options import Options
from spinifex.times import get_gps_week, get_unique_days

# We need to support downloading from the following sources:
# "cddis.nasa.gov": cddis_nasa_gov,
# "chapman.upc.es": chapman_upc_es
# "igsiono.uwm.edu.pl": igsiono_uwm_edu_pl

CENTER_NAMES = {
    "cod",
    "esa",
    "igs",
    "jpl",
    "upc",
    "irt",
    "uqr",
}
DEFAULT_TIME_RESOLUTIONS: dict[str, u.Quantity] = {
    "cod": 1 * u.hour,
    "esa": 2 * u.hour,
    "igs": 2 * u.hour,
    "jpl": 2 * u.hour,
    "upc": 2 * u.hour,
    "irt": 2 * u.hour,
    "uqr": 15 * u.min,  # Chapman only provides 15 minute resolution right now
}


class Servers(str, Enum):
    CDDIS = ("cddis", "https://cddis.nasa.gov/archive/gnss/products/ionex")
    CHAPMAN = ("chapman", "http://chapman.upc.es/tomion/rapid")
    IGSIONO = ("igsiono", "ftp://igs-final.man.olsztyn.pl")

    def __new__(cls, value: str, _url: str) -> Servers:
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj._url = _url  # type: ignore[attr-defined]
        return obj

    @property
    def url(self) -> str:
        return str(self._url)  # type: ignore[attr-defined]

    @classmethod
    def get_url(cls, server: Servers) -> str:
        return str(server.url)  # Retrieve URL given an Enum value


assert set(DEFAULT_TIME_RESOLUTIONS.keys()) == CENTER_NAMES, (
    "Time resolutions must be defined for all analysis centres"
)
NAME_SWITCH_WEEK = 2238  # GPS Week where the naming convention changed

SOLUTION = Literal["final", "rapid"]


def _ftp_download_and_quit(ftp: FTP, file_name: str, output_file: Path) -> None:
    """Download a file from an FTP server and quit the connection.

    Parameters
    ----------
    ftp : FTP
        FTP connection.
    file_name : str
        File name to download.
    output_file : Path
        Output file path.

    Raises
    ------
    e
        If the download fails.
    """
    try:
        with output_file.open("wb") as file_desc:
            ftp.retrbinary(f"RETR {file_name}", file_desc.write)
    except Exception as e:
        output_file.unlink(missing_ok=True)
        raise e
    finally:
        ftp.quit()


async def download_file_ftp(
    url: str,
    output_file: Path,
) -> None:
    """Download a file from a given URL using asyncio.

    Parameters
    ----------
    url : str
        URL to download.
    output_file : Path
        Output file path.
    """
    url_parsed = urlparse(url)
    url_path = Path(url_parsed.path)
    file_name = url_path.name
    directory_name = url_path.parent.as_posix()[1:]  # Remove leading slash

    ftp = FTP(url_parsed.netloc)
    # Anonymous login
    ftp.login()
    ftp.cwd(directory_name)

    await asyncio.to_thread(_ftp_download_and_quit, ftp, file_name, output_file)


async def download_file_http(
    url: str,
    output_file: Path,
    timeout_seconds: int = 30,
    chunk_size: int = 1000,
) -> None:
    """Download a file from a given URL using asyncio.

    Parameters
    ----------
    url : str
        URL to download.
    output_file : Path
        Output file path.
    timeout_seconds : int, optional
        Seconds to wait for request timeout, by default 30
    chunk_size : int, optional
        Chunks of data to download, by default 1000

    Raises
    ------
    IonexError
        If the download times out.
    """
    msg = f"Downloading from {url}"
    logger.info(msg)
    try:
        response = await asyncio.to_thread(requests.get, url, timeout=timeout_seconds)
    except requests.exceptions.Timeout as e:
        msg = "Timed out connecting to server"
        logger.error(msg)
        raise IonexError(msg) from e

    response.raise_for_status()
    with output_file.open("wb") as file_desc:
        for chunk in response.iter_content(chunk_size=chunk_size):
            await asyncio.to_thread(file_desc.write, chunk)


async def download_or_copy_url(
    url: str,
    output_directory: Path | None = None,
    chunk_size: int = 1000,
    timeout_seconds: int = 30,
) -> Path:
    """Download a file from a given URL.

    If the URL is a file URL (i.e. starting with `file://`), it will be copied to the output directory.

    Parameters
    ----------
    url : str
        URL to download.
    output_directory : Path | None, optional
        Output directory, by default None. If None, will default to `ionex_files` in the current working directory.
    chunk_size : int, optional
        Download chunks, by default 1000
    timeout_seconds : int, optional
        Request timeout in seconds, by default 30

    Returns
    -------
    Path
        Output file path

    Raises
    ------
    FileNotFoundError
        If the .netrc file is not found when downloading from CDDIS.
    """
    if output_directory is None:
        output_directory = Path.cwd() / "ionex_files"

    output_directory.mkdir(exist_ok=True)

    url_parsed = urlparse(url)
    url_path = Path(url_parsed.path)
    file_name = url_path.name
    output_file = output_directory / file_name

    if output_file.exists():
        msg = f"File {output_file} already exists. Skipping download."
        logger.info(msg)
        return output_file

    if url_parsed.scheme == "file":
        msg = f"URL scheme {url_parsed.scheme} is not supported"
        logger.info(msg)
        result = await asyncio.to_thread(shutil.copy, url_path, output_file)
        return Path(result)

    if url_parsed.scheme == "ftp":
        await download_file_ftp(url, output_file)
        return output_file

    if url.startswith("https://cddis.nasa.gov"):
        # CDDIS requires a .netrc file to download
        netrc = Path("~/.netrc").expanduser()
        if not netrc.exists():
            msg = "See: https://cddis.nasa.gov/Data_and_Derived_Products/CreateNetrcFile.html"
            logger.error(msg)
            msg = "Please add your NASA Earthdata login credentials to ~/.netrc"
            logger.error(msg)
            raise FileNotFoundError(msg)

    await download_file_http(
        url, output_file, timeout_seconds=timeout_seconds, chunk_size=chunk_size
    )

    return output_file


def new_cddis_format(
    time: Time,
    prefix: str = "cod",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
) -> str:
    """Get the URL for a new IONEX file from CDDIS.

    Parameters
    ----------
    time : Time
        Time to get the URL for.
    prefix : str, optional
        Analysis centre prefix, by default "cod". Must be a supported analysis centre.
    time_resolution : u.Quantity | None, optional
        Time resolution, by default None, will default to the server time resolution.
    url_stem : str | None, optional
        URL steam, by default None, will default to CDDIS.
    solution : SOLUTION, optional
        Solution type, by default "final", must be "final" or "rapid".

    Returns
    -------
    str
        File URL
    """
    # Name Format Since GPS Week 2238
    # WWWW/IGS0OPSTYP_YYYYDDDHHMM_01D_SMP_CNT.INX.gz
    # Code	Meaning
    # WWWW	GPS week
    # TYP	solution type identifier -FIN (Final solution combination)
    # RAP (Rapid solution combination
    # YYYY	4-digit year
    # DDD	3-digit day of year
    # HH	2-digit hour
    # MM	2-digit minute
    # SMP	temporal product sampling resolution
    # CNT	content type -GIM (global ionosphere (TEC) maps)
    # ROT (rate of TEC index maps)
    # .gz	gzip compressed file
    assert time.isscalar, "Only one time is supported"
    if prefix not in CENTER_NAMES:
        msg = f"prefix {prefix} is not supported. Supported prefixes are {CENTER_NAMES}"
        raise IonexError(msg)

    # Parse time resolution
    if time_resolution is None:
        time_resolution = DEFAULT_TIME_RESOLUTIONS.get(prefix)
        if time_resolution is None:
            msg = f"Time resolution not defined for {prefix}"
            raise TimeResolutionError(msg)
        msg = f"Using default time resolution {time_resolution} for {prefix}"
        logger.info(msg)
    if not time_resolution.value.is_integer():
        error = f"Time resolution must be an integer. Got {time_resolution}"
        raise TimeResolutionError(error)
    if time_resolution.to(u.min).value % 15 != 0:
        msg = f"Time resolution on CDDIS is multiples 15 minutes. Please check the time resolution ({time_resolution})."
        logger.warning(msg)
    if time_resolution < 1 * u.hour:
        time_resolution = time_resolution.to(u.min)
    else:
        time_resolution = time_resolution.to(u.hour)
    time_res_str = (
        f"{int(time_resolution.value):02d}{str(time_resolution.unit)[0].upper()}"
    )

    if solution == "final":
        solution_str = "FIN"
    elif solution == "rapid":
        solution_str = "RAP"
    else:
        msg = f"Solution {solution} is not supported. Supported solutions are ['final', 'rapid']"  # type: ignore[unreachable]
        raise IonexError(msg)

    dtime: datetime = time.to_datetime()
    doy = time.datetime.timetuple().tm_yday
    prefix_formatted = prefix.upper()

    # WWWW/IGS0OPSTYP_YYYYDDDHHMM_01D_SMP_CNT.INX.gz
    file_name = f"{prefix_formatted}0OPS{solution_str}_{dtime.year:03d}{doy:03d}0000_01D_{time_res_str}_GIM.INX.gz"
    directory = f"{dtime.year:04d}/{doy:03d}"

    if url_stem is None:
        url_stem = Servers.get_url(Servers.CDDIS)

    return f"{url_stem}/{directory}/{file_name}"


def old_cddis_format(
    time: Time,
    prefix: str = "cod",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
) -> str:
    """Get the URL for an old IONEX file from CDDIS.

    Parameters
    ----------
    time : Time
        Time to get the URL for.
    prefix : str, optional
        Analysis centre prefix, by default "cod"
    time_resolution : u.Quantity | None, optional
        Time resolution, by default None, will default to the server time resolution.
    url_stem : str | None, optional
        URL steam, by default None, will default to CDDIS.
    solution : SOLUTION, optional
        Solution type, by default "final", must be "final" or "rapid".

    Returns
    -------
    str
        File URL

    Raises
    ------
    IonexError
        If the prefix is not a supported analysis centre.
    """
    # Name Format Before GPS Week 2237
    # YYYY/DDD/AAAgDDD#.YYi.Z - Vertical total electron content (TEC) maps
    # Code	Meaning
    # YYYY	4-digit year
    # DDD	3-digit day of year
    # AAA	Analysis center name
    # #	file number for the day, typically 0
    # YY	2-digit year
    # .Z	Unix compressed file
    if time_resolution is not None:
        msg = f"Time resolution is not used by the old CDDIS format (got `{time_resolution}`). Ignoring."
        logger.warning(msg)

    assert time.isscalar, "Only one time is supported"
    if prefix not in CENTER_NAMES:
        msg = f"prefix {prefix} is not supported. Supported prefixes are {CENTER_NAMES}"
        raise IonexError(msg)

    if solution == "final":
        prefix_str = prefix
    elif solution == "rapid":
        prefix_str = prefix[:-1] + "r"
    else:
        msg = f"Solution {solution} is not supported. Supported solutions are ['final', 'rapid']"  # type: ignore[unreachable]
        raise IonexError(msg)

    dtime: datetime = time.to_datetime()
    doy = time.datetime.timetuple().tm_yday
    # YYYY/DDD/AAAgDDD#.YYi.Z
    yy = f"{dtime.year:02d}"[-2:]
    file_name = f"{prefix_str}g{doy:03d}0.{yy}i.Z"
    directory_name = f"{dtime.year:04d}/{doy:03d}"
    if url_stem is None:
        url_stem = Servers.get_url(Servers.CDDIS)

    return f"{url_stem}/{directory_name}/{file_name}"


async def download_from_cddis(
    times: Time,
    prefix: str = "cod",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
    output_directory: Path | None = None,
) -> list[Path]:
    """Download IONEX files from CDDIS.

    Parameters
    ----------
    times : Time
        Times to download for.
    prefix : str, optional
        Analysis centre prefix, by default "cod"
    time_resolution : u.Quantity | None, optional
        Time resolution, by default None, will default to the server time resolution.
    url_stem : str | None, optional
        URL steam, by default None, will default to CDDIS.
    solution : SOLUTION, optional
        Solution type, by default "final", must be "final" or "rapid".
    output_directory : Path | None, optional
        Output directory path, by default None, will default to `ionex_files` in the current working directory.

    Returns
    -------
    list[Path]
        List of downloaded files
    """
    unique_days: Time = get_unique_days(times)

    coros = []
    for day in unique_days:
        if get_gps_week(day) >= NAME_SWITCH_WEEK:
            formatter = new_cddis_format
        else:
            formatter = old_cddis_format
        url = formatter(
            day,
            prefix=prefix,
            url_stem=url_stem,
            time_resolution=time_resolution,
            solution=solution,
        )
        coros.append(download_or_copy_url(url, output_directory=output_directory))

    return await asyncio.gather(*coros)


def chapman_format(
    time: Time,
    prefix: str = "uqr",
    url_stem: str | None = None,
) -> str:
    """Format a URL for an IRTG file from Chapman.

    Parameters
    ----------
    time : Time
        Time to get the URL for.
    prefix : str, optional
        Ionex prefix, by default "irt"
    url_stem : str | None, optional
        URL stem, by default None

    Returns
    -------
    str
        File URL
    """

    assert time.isscalar, "Only one time is supported"

    dtime: datetime = time.to_datetime()
    doy = int(time.datetime.timetuple().tm_yday)
    yy = f"{dtime.year:02d}"[-2:]
    # e.g. uqrg3260.96i.Z
    file_name = f"{prefix}g{doy:03d}0.{yy}i.Z"
    # e.g. 1996/326_961121.15min/
    # Hard coded to 15 minute resolution - Chapman only provides 15 minute resolution right now
    # Could be updated to be more flexible in the future
    directory_name = (
        f"{dtime.year:04d}/{doy:03d}_{yy}{dtime.month:02d}{dtime.day:02d}.15min"
    )
    if url_stem is None:
        url_stem = Servers.get_url(Servers.CHAPMAN)

    return f"{url_stem}/{directory_name}/{file_name}"


async def download_from_chapman(
    times: Time,
    prefix: str = "uqr",
    url_stem: str | None = None,
    output_directory: Path | None = None,
) -> list[Path]:
    """Download IONEX files from Chapman.

    Parameters
    ----------
    times : Time
        Times to download for.
    prefix : str, optional
        Ionex prefix, by default "irt"
    url_stem : str | None, optional
        URL stem, by default None
    output_directory : Path | None, optional
        Output directory path, by default None

    Returns
    -------
    list[Path]
        List of downloaded file paths

    Raises
    ------
    IonexError
        If the date is before 2021/024.
    """
    unique_days: Time = get_unique_days(times)
    for day in unique_days:
        # Chapman goes back to 326th day of 1996 == 21/11/1996
        _chapman_start = datetime(1996, 11, 21)
        if day.datetime < _chapman_start:
            msg = f"Date {day.datetime} is before {_chapman_start}. Chapman only goes back to this date."
            raise IonexError(msg)

    coros = []
    for day in unique_days:
        url = chapman_format(day, prefix=prefix, url_stem=url_stem)
        coros.append(download_or_copy_url(url, output_directory=output_directory))

    return await asyncio.gather(*coros)


def igsiono_format(
    time: Time,
    prefix: str = "igs",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
) -> str:
    """Format a URL for an IGS IONO file from IGSIONO.

    Parameters
    ----------
    time : Time
        Time to get the URL for.
    prefix : str, optional
        Ionex prefix, by default "igs"
    time_resolution : u.Quantity | None, optional
        Ionex resolution, by default None
    url_stem : str | None, optional
        URL stem, by default None
    solution : SOLUTION, optional
        Type of solution, by default "final"

    Returns
    -------
    str
        File URL
    """
    assert time.isscalar, "Only one time is supported"

    dtime: datetime = time.to_datetime()
    doy = time.datetime.timetuple().tm_yday
    yy = f"{dtime.year:02d}"[-2:]
    directory_name = f"pub/gps_data/GPS_IONO/cmpcmb/{yy}{doy:03d}"
    if url_stem is None:
        url_stem = Servers.get_url(Servers.IGSIONO)

    # File name matches CDDIS format
    if get_gps_week(time) >= NAME_SWITCH_WEEK:
        formatter = new_cddis_format
    else:
        formatter = old_cddis_format
    file_name = Path(
        formatter(
            time,
            prefix=prefix,
            url_stem=url_stem,
            time_resolution=time_resolution,
            solution=solution,
        )
    ).name
    return f"{url_stem}/{directory_name}/{file_name}"


async def download_from_igsiono(
    times: Time,
    prefix: str = "igs",
    time_resolution: u.Quantity | None = None,
    url_stem: str | None = None,
    solution: SOLUTION = "final",
    output_directory: Path | None = None,
) -> list[Path]:
    """Download IONEX files from IGSIONO.

    Parameters
    ----------
    times : Time
        Times to download for.
    prefix : str, optional
        URL prefix, by default "igs"
    time_resolution : u.Quantity | None, optional
        Ionex resolution, by default None
    url_stem : str | None, optional
        URL stem, by default None
    solution : SOLUTION, optional
        Type of solution, by default "final"
    output_directory : Path | None, optional
        Output directory path, by default None

    Returns
    -------
    list[Path]
        List of downloaded file paths
    """
    unique_days: Time = get_unique_days(times)

    coros = []
    for day in unique_days:
        url = igsiono_format(
            day,
            prefix=prefix,
            time_resolution=time_resolution,
            url_stem=url_stem,
            solution=solution,
        )
        coros.append(download_or_copy_url(url, output_directory=output_directory))

    return await asyncio.gather(*coros)


class IonexOptions(Options):
    """Options for ionex model"""

    server: Servers = Field(
        Servers.CHAPMAN, description="Server to download ionex files from"
    )
    prefix: str = Field("uqr", description="Analysis centre prefix")
    url_stem: str | None = Field(None, description="URL stem")
    time_resolution: u.Quantity | None = Field(
        None, description="Time resolution for ionex files"
    )
    solution: SOLUTION = Field("final", description="Solution type")
    output_directory: Path | None = Field(
        None, description="Output directory for ionex files"
    )


async def _download_ionex_coro(
    times: Time,
    options: IonexOptions | None = None,
) -> list[Path]:
    """Download IONEX files from a server.

    Parameters
    ----------
    times : Time
        Times to download for.
    options : IonexOptions
        Options for the ionex model.

    Returns
    -------
    list[Path]
        List of downloaded files

    Raises
    ------
    IonexError
        If the server is not supported.
    NotImplementedError
        If the server is not implemented yet.
    """
    if options is None:
        options = IonexOptions()

    # TODO: Consider refactor to pass around the options object
    if options.server == Servers.CDDIS:
        return await download_from_cddis(
            times,
            prefix=options.prefix,
            url_stem=options.url_stem,
            time_resolution=options.time_resolution,
            solution=options.solution,
            output_directory=options.output_directory,
        )
    if options.server == Servers.CHAPMAN:
        if options.prefix != "uqr":
            msg = f"Chapman server primarily supports uqr prefix. You provided {options.prefix}, this may fail"
            logger.warning(msg)

        return await download_from_chapman(
            times,
            prefix=options.prefix,
            url_stem=options.url_stem,
            output_directory=options.output_directory,
        )

    # Must be IGSIONO
    assert options.server == Servers.IGSIONO, "Server not supported"
    if options.prefix != "igs":
        msg = f"IGS IONO server primarily supports igs prefix. You provided {options.prefix}, this may fail"
        logger.warning(msg)
    return await download_from_igsiono(
        times,
        prefix=options.prefix,
        time_resolution=options.time_resolution,
        url_stem=options.url_stem,
        solution=options.solution,
        output_directory=options.output_directory,
    )


async def download_ionex_coro(
    times: Time,
    server: str = "chapman",
    prefix: str = "uqr",
    url_stem: str | None = None,
    time_resolution: u.Quantity | None = None,
    solution: SOLUTION = "final",
    output_directory: Path | None = None,
) -> list[Path]:
    """Download IONEX files from a server.

    Parameters
    ----------
    times : Time
        Times to download for.
    server : str
        Server to download from, by default "cddis". Must be a supported server.
    prefix : str, optional
        Analysis centre prefix, by default "cod". Must be a supported analysis centre.
    url_stem : str | None, optional
        URL stem, by default None, will default to the server URL.
    time_resolution : u.Quantity | None, optional
        Time resolution, by default None, will default to the server time resolution.
    solution : SOLUTION, optional
        Solution type, by default "final", must be "final" or "rapid".
    output_directory : Path | None, optional
        Output directory path, by default None, will default to `ionex_files` in the current working directory.

    Returns
    -------
    list[Path]
        List of downloaded files

    Raises
    ------
    IonexError
        If the server is not supported.
    NotImplementedError
        If the server is not implemented yet.
    """

    options = IonexOptions(
        server=server,
        prefix=prefix,
        url_stem=url_stem,
        time_resolution=time_resolution,
        solution=solution,
        output_directory=output_directory,
    )

    return await _download_ionex_coro(times, options)


# Create synchronous wrapper of download_ionex_coro
download_ionex = sync_wrapper(download_ionex_coro)
_download_ionex = sync_wrapper(_download_ionex_coro)



from __future__ import annotations

import argparse
from concurrent.futures import Future, ThreadPoolExecutor
from functools import partial
from pathlib import Path
from typing import Any, Literal, NamedTuple

import astropy.units as u
import numpy as np
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from numpy.typing import NDArray

from spinifex import h5parm_tools
from spinifex.get_dtec import DTEC, _get_dtec_from_skycoord
from spinifex.get_rm import (
    DEFAULT_IONO_HEIGHT,
    RM,
    _get_rm_from_skycoord,
)
from spinifex.ionospheric import ModelDensityFunction, ionospheric_models
from spinifex.ionospheric.models import IonoOptions, O, parse_iono_kwargs
from spinifex.logger import logger
from spinifex.magnetic import MagneticFieldFunction, magnetic_models

try:
    from casacore.tables import table as _casacore_table
    from casacore.tables import taql
except ImportError as e:
    MSG = "casacore is not installed! To operate on MeasurementSets, install spinifex[casacore]."
    raise ImportError(MSG) from e

# Disable acknowledgement from opening casacore tables
table = partial(_casacore_table, ack=False)


class MsMetaData(NamedTuple):
    """Metadata from a Measurement Set"""

    times: Time
    locations: EarthLocation
    station_names: list[str]
    name: str
    source: SkyCoord


def get_metadata_from_ms(ms_path: Path) -> MsMetaData:
    """open measurement set and get metadata from it

    Parameters
    ----------
    ms_path : Path
        measurement set

    Returns
    -------
    MsMetaData
        object with metadata
    """
    timerange = list(
        taql("select gmin(TIME_CENTROID), gmax(TIME_CENTROID) from $ms_path")[
            0
        ].values()
    )
    with table(ms_path.as_posix()) as my_ms:
        timestep = my_ms.getcell("INTERVAL", 0)
        times = Time(
            np.arange(timerange[0], timerange[1] + 0.5 * timestep, timestep)
            / (24 * 3600),
            format="mjd",
        )
        pointing = table(my_ms.getkeyword("FIELD")).getcell("PHASE_DIR", 0)[0]
        stations = table(my_ms.getkeyword("ANTENNA")).getcol("NAME")
        station_pos = table(my_ms.getkeyword("ANTENNA")).getcol("POSITION")
        locations = EarthLocation.from_geocentric(*station_pos.T, unit=u.m)
        return MsMetaData(
            times=times,
            locations=locations,
            station_names=stations,
            name=ms_path.as_posix(),
            source=SkyCoord(pointing[0] * u.rad, pointing[1] * u.rad),
        )


def get_columns_from_ms(ms_path: Path) -> list[str]:
    """Get the columns from a MeasurementSet"""
    with table(ms_path.as_posix()) as my_ms:
        return list(my_ms.colnames())


def get_average_location(location: EarthLocation) -> EarthLocation:
    # TODO; implement correctly in NE plane
    """Get first location from N locations


    Parameters
    ----------
    location : EarthLocation
        N locations

    Returns
    -------
    EarthLocation
        first location
    """
    logger.warning("Using first location from list of locations - not a true average!")
    return location[0]


def get_rm_from_ms(
    ms_path: Path,
    timestep: u.Quantity | None = None,
    use_stations: list[int] | list[str] | Literal["all", "average"] = "average",
    height_array: NDArray[np.float64] = DEFAULT_IONO_HEIGHT,
    iono_model: ModelDensityFunction[IonoOptions] = ionospheric_models.ionex,
    magnetic_model: MagneticFieldFunction = magnetic_models.ppigrf,
    **iono_kwargs: Any,
) -> dict[str, RM]:
    """Get rotation measures for a measurement set

    Parameters
    ----------
    ms_path : Path
        measurement set
    timestep : u.Quantity | None, optional
        only calculate rotation measure every timestep, by default None
    use_stations : list[int  |  str] | None, optional
        list of stations (index or name) to use,
        if None use first of the measurement set, by default None
    height_array : NDArray[np.float64], optional
        array of ionospheric altitudes, by default DEFAULT_IONO_HEIGHT
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    magnetic_model : MagneticFieldFunction, optional
        geomagnetic model, by default magnetic_models.ppigrf
    iono_kwargs : dict, optional
        arguments for the ionospheric model, by default None

    Returns
    -------
    dict[str, RM]
        dictionary with RM object per station
    """
    # TODO: Expose and create iono_options
    iono_options = parse_iono_kwargs(iono_model, **iono_kwargs)
    return _get_iono_from_ms(
        ms_path=ms_path,
        iono_type="rm",
        timestep=timestep,
        use_stations=use_stations,
        height_array=height_array,
        iono_model=iono_model,
        magnetic_model=magnetic_model,
        iono_options=iono_options,
    )


def get_dtec_from_ms(
    ms_path: Path,
    timestep: u.Quantity | None = None,
    use_stations: list[int] | list[str] | Literal["all", "average"] = "average",
    height_array: NDArray[np.float64] = DEFAULT_IONO_HEIGHT,
    iono_model: ModelDensityFunction[IonoOptions] = ionospheric_models.ionex,
    **iono_kwargs: Any,
) -> dict[str, NDArray]:
    """Get rotation measures for a measurement set

    Parameters
    ----------
    ms_path : Path
        measurement set
    timestep : u.Quantity | None, optional
        only calculate rotation measure every timestep, by default None
    use_stations : list[int  |  str] | None, optional
        list of stations (index or name) to use,
        if None use first of the measurement set, by default None
    height_array : NDArray[np.float64], optional
        array of ionospheric altitudes, by default DEFAULT_IONO_HEIGHT
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    iono_options : OptionType | None, optional
        arguments for the ionospheric model, by default None

    Returns
    -------
    dict[str, NDArray]
        dictionary with electron_density_profiles per station
    """
    iono_options = parse_iono_kwargs(iono_model, **iono_kwargs)
    return _get_iono_from_ms(
        ms_path=ms_path,
        iono_type="dtec",
        timestep=timestep,
        use_stations=use_stations,
        height_array=height_array,
        iono_model=iono_model,
        iono_options=iono_options,
    )


def _get_iono_from_ms(
    ms_path: Path,
    iono_type: Literal["dtec", "rm"],
    timestep: u.Quantity | None = None,
    use_stations: list[int] | list[str] | Literal["all", "average"] = "average",
    height_array: NDArray[np.float64] = DEFAULT_IONO_HEIGHT,
    iono_model: ModelDensityFunction[O] = ionospheric_models.ionex,
    magnetic_model: MagneticFieldFunction = magnetic_models.ppigrf,
    iono_options: O | None = None,
) -> dict[str, RM] | dict[str, NDArray]:
    """Get ionospheric values for a measurement set

    Parameters
    ----------
    ms_path : Path
        measurement set
    iono_type : Literal["dtec", "rm"]
        type of ionospheric value to calculate
    timestep : u.Quantity | None, optional
        only calculate rotation measure every timestep, by default None
    use_stations : list[int  |  str] | None, optional
        list of stations (index or name) to use,
        if None use first of the measurement set, by default None
    height_array : NDArray[np.float64], optional
        array of ionospheric altitudes, by default DEFAULT_IONO_HEIGHT
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    magnetic_model : MagneticFieldFunction, optional
        geomagnetic model, by default magnetic_models.ppigrf
    iono_options : OptionType | None, optional
        arguments for the ionospheric model, by default None

    Returns
    -------
    dict[str, RM] | dict[str, NDArray]
        dictionary with ionospheric values
    """

    if iono_type not in ["dtec", "rm"]:
        msg = f"iono_type should be 'dtec' or 'rm'. Got {iono_type}"
        raise ValueError(msg)

    get_func = (
        _get_dtec_from_skycoord
        if iono_type == "dtec"
        # Set the magnetic model here for get_rm
        # all other arguments are the same
        else partial(_get_rm_from_skycoord, magnetic_model=magnetic_model)
    )

    result_dict: dict[str, RM | DTEC] = {}

    ms_metadata = get_metadata_from_ms(ms_path)
    if timestep is not None:
        # dtime = ms_metadata.times[1].mjd - ms_metadata.times[0].mjd
        dtime_in_days = timestep.to(u.hr).value / 24
        times = Time(
            np.arange(
                ms_metadata.times[0].mjd - 0.5 * dtime_in_days,
                ms_metadata.times[-1].mjd + 0.5 * dtime_in_days,
                dtime_in_days,
            ),
            format="mjd",
        )
    else:
        times = ms_metadata.times

    if isinstance(use_stations, str):
        if use_stations == "average":
            location = get_average_location(ms_metadata.locations)
            result_dict["average_station_pos"] = get_func(  # type: ignore[operator]
                loc=location,
                times=times,
                source=ms_metadata.source,
                height_array=height_array,
                iono_model=iono_model,
                iono_options=iono_options,
            )
            return result_dict
        if use_stations == "all":
            station_list: list[str] | list[int] = ms_metadata.station_names
        else:
            msg = f"`use_stations` should be a list of stations, or string literal 'average' or 'all'. Got {use_stations}"  # type: ignore[unreachable]
            raise ValueError(msg)

    elif isinstance(use_stations, list):
        station_list = use_stations
    else:
        msg = f"`use_stations` should be a list of stations, or string literal 'average' or 'all'. Got {use_stations}"  # type: ignore[unreachable]
        raise ValueError(msg)
    # get rm per station
    logger.info("Getting DTEC per station")

    # get rm per station
    # Submit jobs to ThreadPoolExecutor
    # Gets Future objects for each station
    future_dict: dict[str, Future[RM | DTEC]] = {}
    key_names: list[str] = []

    # Execute the first station in serial
    # Avoids race conidition to download ionex files

    stat = station_list[0]
    istat = ms_metadata.station_names.index(stat) if isinstance(stat, str) else stat
    key_name = ms_metadata.station_names[istat]
    result_dict[key_name] = get_func(  # type: ignore[operator]
        loc=ms_metadata.locations[istat],
        times=times,
        source=ms_metadata.source,
        height_array=height_array,
        iono_model=iono_model,
        iono_options=iono_options,
    )

    with ThreadPoolExecutor() as executor:
        # Skip the first station
        for stat in station_list[1:]:
            istat = (
                ms_metadata.station_names.index(stat) if isinstance(stat, str) else stat
            )
            key_name = ms_metadata.station_names[istat]
            key_names.append(key_name)
            future_dict[key_name] = executor.submit(
                get_func,  # type: ignore[arg-type]
                loc=ms_metadata.locations[istat],
                times=times,
                source=ms_metadata.source,
                height_array=height_array,
                iono_model=iono_model,
                iono_options=iono_options,
            )

    # Get concrete results from futures
    # Resolve Future objects to get the results
    for key_name in key_names:
        result_dict[key_name] = future_dict[key_name].result()
    return result_dict


def cli_get_rm_h5parm_from_ms(args: argparse.Namespace) -> None:
    ms_path: Path = Path(args.ms)
    h5parm_path: str = args.h5parm
    solset_name: str | None = args.solset_name
    soltab_name: str | None = args.soltab_name
    add_to_existing_solset: bool = args.add_to_existing_solset

    rm_dict = get_rm_from_ms(
        ms_path=ms_path,
        use_stations="all",
    )
    h5parm_tools.write_rm_to_h5parm(
        rm_dict,
        h5parm_path,
        solset_name=solset_name,
        soltab_name=soltab_name,
        add_to_existing_solset=add_to_existing_solset,
    )


def cli_get_dtec_h5parm_from_ms(args: argparse.Namespace) -> None:
    ms_path: Path = Path(args.ms)
    h5parm_path: str = args.h5parm
    solset_name: str | None = args.solset_name
    soltab_name: str | None = args.soltab_name
    add_to_existing_solset: bool = args.add_to_existing_solset

    dtec = get_dtec_from_ms(
        ms_path=ms_path,
        use_stations="all",
    )
    h5parm_tools.write_tec_to_h5parm(
        dtec,
        h5parm_path,
        solset_name=solset_name,
        soltab_name=soltab_name,
        add_to_existing_solset=add_to_existing_solset,
    )



#  Copyright (C) 2023 ASTRON (Netherlands Institute for Radio Astronomy)
#  SPDX-License-Identifier: Apache-2.0
"""Testing of the main module"""

from __future__ import annotations

from astropy.utils import iers

iers.conf.auto_download = False

import astropy.units as u
import h5py
import numpy as np
import pytest
from astropy.coordinates import EarthLocation
from astropy.time import Time
from spinifex import h5parm_tools as h5pt
from spinifex.get_rm import RM


def test_create_h5parm(tmpdir):
    """test that h5parm is created"""
    h5parm_name = f"{tmpdir}/test.h5"
    h5pt.create_empty_h5parm(h5parm_name)
    with h5py.File(h5parm_name, "a") as h5parm:
        solset = h5pt.create_solset(h5parm, "mysolset")
        assert "mysolset" in h5parm
        station_names = ["station1"]
        station_pos = [[0, 0, 0]]
        h5pt.add_antenna_info(
            solset=solset, antenna_names=station_names, antenna_pos=station_pos
        )
        assert solset["antenna"]["name"][0].decode() == "station1"
        source_names = ["source1"]
        source_dir = [[0, 0]]
        h5pt.add_source_info(
            solset=solset, source_names=source_names, source_dirs=source_dir
        )
        assert solset["source"]["name"][0].decode() == "source1"
        axes_values = {}
        axes_values["axis1"] = [1, 2]
        axes_values["axis2"] = [3, 4, 5]
        soltab_axes = ["axis1", "axis2"]
        val = np.ones((2, 3))
        weight = np.ones(val.shape, dtype=bool)
        h5pt.add_soltab(
            solset=solset,
            soltab_type="rotationmeasure",
            val=val,
            weight=weight,
            soltab_axes=soltab_axes,
            axes_values=axes_values,
        )
        assert "rotationmeasure000" in solset
        assert solset["rotationmeasure000"]["val"].shape == (2, 3)


def test_write_rm_h5parm(tmpdir):
    """test that writing rotation measures does not crash"""
    h5parm_name = f"{tmpdir}/test.h5"
    rms = {}
    lon = 6.367 * u.deg
    lat = 52.833 * u.deg
    dwingeloo = EarthLocation(lon=lon, lat=lat, height=0 * u.km)
    rm = RM(
        rm=np.arange(0, 1, 0.1),
        times=Time("2025-02-08") + np.arange(10) * u.hr,
        b_parallel=np.zeros((10, 1)),
        electron_density=np.zeros((10, 1)),
        height=np.array([350.0]),
        azimuth=np.zeros((10,)),
        elevation=np.zeros((10,)),
        loc=dwingeloo,
    )
    stations = ["CS001LBA", "CS002LBA"]
    for stat in stations:
        rms[stat] = rm
    h5pt.write_rm_to_h5parm(rms, h5parm_name=h5parm_name)
    # check sol000 exists, check sol000/rotationmeasure000 exists
    with h5py.File(h5parm_name) as myh5:
        assert "sol000" in myh5
        assert "sol000/rotationmeasure000" in myh5
    h5pt.write_rm_to_h5parm(rms, h5parm_name=h5parm_name)
    # check sol001 exists, check sol001/rotationmeasure000 exists
    with h5py.File(h5parm_name) as myh5:
        assert "sol001" in myh5
        assert "sol001/rotationmeasure000" in myh5
    with pytest.raises(RuntimeError, match="sol000 already exists"):
        h5pt.write_rm_to_h5parm(rms, h5parm_name=h5parm_name, solset_name="sol000")
        # check runtime error
    h5pt.write_rm_to_h5parm(
        rms, h5parm_name=h5parm_name, solset_name="sol000", add_to_existing_solset=True
    )
    # check sol000 exists, check sol000/rotationmeasure001 exists
    with h5py.File(h5parm_name) as myh5:
        assert "sol000" in myh5
        assert "sol000/rotationmeasure001" in myh5
    with pytest.raises(RuntimeError, match="rotationmeasure000 already exists"):
        h5pt.write_rm_to_h5parm(
            rms,
            h5parm_name=h5parm_name,
            solset_name="sol000",
            soltab_name="rotationmeasure000",
            add_to_existing_solset=True,
        )
        # check runtime error
    h5pt.write_rm_to_h5parm(
        rms,
        h5parm_name=h5parm_name,
        solset_name="sol000",
        soltab_name="RMextract",
        add_to_existing_solset=True,
    )
    # check sol000 exists, check sol000/RMextract exists
    with h5py.File(h5parm_name) as myh5:
        assert "sol000" in myh5
        assert "sol000/RMextract" in myh5
    h5pt.write_rm_to_h5parm(rms, h5parm_name=h5parm_name, solset_name="sol042")
    # check sol042 exists, check sol042/rotationmeasure000 exists
    with h5py.File(h5parm_name) as myh5:
        assert "sol042" in myh5
        assert "sol042/rotationmeasure000" in myh5
    h5pt.write_rm_to_h5parm(
        rms,
        h5parm_name=h5parm_name,
        soltab_name="RMextract",
        add_to_existing_solset=True,
    )
    # check sol002 exists, check sol002/RMextract exists
    with h5py.File(h5parm_name) as myh5:
        assert "sol002" in myh5
        assert "sol002/RMextract" in myh5



"""Module to calculate rotation measures"""

from __future__ import annotations

from typing import Any, NamedTuple

import astropy.units as u
import numpy as np
from astropy.coordinates import AltAz, EarthLocation, SkyCoord
from astropy.time import Time
from numpy.typing import NDArray

from spinifex.geometry import IPP, get_ipp_from_altaz, get_ipp_from_skycoord
from spinifex.ionospheric import ModelDensityFunction, ionospheric_models
from spinifex.ionospheric.models import (
    O,
    parse_iono_kwargs,
    parse_iono_model,
)
from spinifex.logger import logger
from spinifex.magnetic import MagneticFieldFunction, magnetic_models
from spinifex.magnetic.models import parse_magnetic_model

DEFAULT_IONO_HEIGHT = np.array([450.0]) * u.km


class RM(NamedTuple):
    """object with all rotation measures"""

    rm: NDArray[Any]
    """rotation measures"""
    times: Time
    """time axis"""
    b_parallel: NDArray[Any]
    """parallel magnetic field"""
    electron_density: NDArray[Any]
    """electron content"""
    height: NDArray[Any]
    """array of altitudes (km)"""
    azimuth: NDArray[Any]
    """array of azimuths (degrees)"""
    elevation: NDArray[Any]
    """array of elevation (degrees)"""
    loc: EarthLocation
    """observer location"""


def _get_rm(
    ipp: IPP,
    iono_model: ModelDensityFunction[O] = ionospheric_models.ionex,
    magnetic_model: MagneticFieldFunction = magnetic_models.ppigrf,
    iono_options: O | None = None,
) -> RM:
    """Get the rotation measures for a given set of ionospheric piercepoints

    Parameters
    ----------
    ipp : IPP
        ionospheric piercepoints
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    magnetic_model : MagneticFieldFunction, optional
        geomagnetic model, by default magnetic_models.ppigrf
    iono_options : OptionType, optional
        options for the ionospheric model, by default None

    Returns
    -------
    RM
        rotation measures object
    """
    logger.info("Calculating rotation measure")
    density_profile = iono_model(ipp=ipp, options=iono_options)
    magnetic_profile = magnetic_model(ipp=ipp)
    b_field_to_rm = -2.62e-6  # TODO: What are the units of this constant?
    rm = np.sum(
        b_field_to_rm * density_profile * magnetic_profile.to(u.nT).value * ipp.airmass,
        axis=1,
    )
    return RM(
        rm=rm,
        times=ipp.times,
        b_parallel=magnetic_profile,
        electron_density=density_profile,
        height=ipp.loc.height.to(u.km).value,
        azimuth=ipp.altaz.az.deg,
        elevation=ipp.altaz.alt.deg,
        loc=ipp.station_loc,
    )


def get_average_rm(rm: RM) -> RM:
    profile_weights = np.sum(rm.electron_density, axis=1, keepdims=True)
    return RM(
        rm=rm.rm.mean(),
        times=rm.times.mean(),
        b_parallel=np.sum(
            rm.b_parallel * rm.electron_density / profile_weights, axis=1
        ).mean(),
        electron_density=profile_weights.mean(),
        height=np.sum(rm.height * rm.electron_density / profile_weights, axis=1).mean(),
        azimuth=np.degrees(np.angle(np.sum(np.exp(1.0j * np.radians(rm.azimuth))))),
        elevation=rm.elevation.mean(),
        loc=rm.loc,
    )


def _get_rm_from_altaz(
    loc: EarthLocation,
    altaz: AltAz,
    height_array: u.Quantity = DEFAULT_IONO_HEIGHT,
    iono_model: ModelDensityFunction[O] = ionospheric_models.ionex,
    magnetic_model: MagneticFieldFunction = magnetic_models.ppigrf,
    iono_options: O | None = None,
) -> RM:
    """get rotation measures for user defined altaz coordinates

    Parameters
    ----------
    loc : EarthLocation
        observer location
    altaz : AltAz
        altaz coordinates
    height_array : u.Quantity, optional
        altitudes, by default default_height
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    magnetic_model : MagneticFieldFunction, optional
        geomagnetic model, by default magnetic_models.ppigrf
    iono_kwargs : dict
        keyword arguments for the ionospheric model

    Returns
    -------
    RM
        rotation measure object
    """
    ipp = get_ipp_from_altaz(loc=loc, altaz=altaz, height_array=height_array)
    return _get_rm(
        ipp=ipp,
        iono_model=iono_model,
        magnetic_model=magnetic_model,
        iono_options=iono_options,
    )


def get_rm_from_altaz(
    loc: EarthLocation,
    altaz: AltAz,
    height_array: u.Quantity = DEFAULT_IONO_HEIGHT,
    iono_model_name: str = "ionex",
    magnetic_model_name: str = "ppigrf",
    **iono_kwargs: Any,
) -> RM:
    """get rotation measures for user defined altaz coordinates

    Parameters
    ----------
    loc : EarthLocation
        observer location
    altaz : AltAz
        altaz coordinates
    height_array : u.Quantity, optional
        altitudes, by default default_height
    iono_model_name : str, optional
        ionospheric model name, by default "ionex". Must be a supported ionospheric model.
    magnetic_model_name : str, optional
        geomagnetic model name, by default "ppigrf". Must be a supported geomagnetic model.
    iono_kwargs : dict
        keyword arguments for the ionospheric model

    Returns
    -------
    RM
        rotation measure object
    """
    iono_model = parse_iono_model(iono_model_name)
    iono_options = parse_iono_kwargs(iono_model=iono_model, **iono_kwargs)
    magnetic_model = parse_magnetic_model(magnetic_model_name)
    return _get_rm_from_altaz(
        loc=loc,
        altaz=altaz,
        height_array=height_array,
        iono_model=iono_model,
        magnetic_model=magnetic_model,
        iono_options=iono_options,
    )


def _get_rm_from_skycoord(
    loc: EarthLocation,
    times: Time,
    source: SkyCoord,
    height_array: u.Quantity = DEFAULT_IONO_HEIGHT,
    iono_model: ModelDensityFunction[O] = ionospheric_models.ionex,
    magnetic_model: MagneticFieldFunction = magnetic_models.ppigrf,
    iono_options: O | None = None,
) -> RM:
    """get rotation measures for user defined times and source coordinate

    Parameters
    ----------
    loc : EarthLocation
        observer location
    times : Time
        times
    source : SkyCoord
        coordinates of the source
    height_array : NDArray, optional
        altitudes, by default default_height
    iono_model : ModelDensityFunction, optional
        ionospheric model, by default ionospheric_models.ionex
    magnetic_model : MagneticFieldFunction, optional
        geomagnetic model, by default magnetic_models.ppigrf
    iono_options : IonoOptions, optional
        options for the ionospheric model, by default None


    Returns
    -------
    RM
        rotation measure object
    """

    ipp = get_ipp_from_skycoord(
        loc=loc, times=times, source=source, height_array=height_array
    )
    return _get_rm(
        ipp=ipp,
        iono_model=iono_model,
        magnetic_model=magnetic_model,
        iono_options=iono_options,
    )


def get_rm_from_skycoord(
    loc: EarthLocation,
    times: Time,
    source: SkyCoord,
    height_array: u.Quantity = DEFAULT_IONO_HEIGHT,
    iono_model_name: str = "ionex",
    magnetic_model_name: str = "ppigrf",
    **iono_kwargs: Any,
) -> RM:
    """get rotation measures for user defined times and source coordinate

    Parameters
    ----------
    loc : EarthLocation
        observer location
    times : Time
        times
    source : SkyCoord
        coordinates of the source
    height_array : NDArray, optional
        altitudes, by default default_height
    iono_model_name : str, optional
        ionospheric model name, by default "ionex". Must be a supported ionospheric model.
    magnetic_model_name : str, optional
        geomagnetic model name, by default "ppigrf". Must be a supported geomagnetic model.
    iono_kwargs : dict
        keyword arguments for the ionospheric model


    Returns
    -------
    RM
        rotation measure object
    """
    iono_model = parse_iono_model(iono_model_name)
    iono_options = parse_iono_kwargs(iono_model=iono_model, **iono_kwargs)
    magnetic_model = parse_magnetic_model(magnetic_model_name)
    return _get_rm_from_skycoord(
        loc=loc,
        times=times,
        source=source,
        height_array=height_array,
        iono_model=iono_model,
        magnetic_model=magnetic_model,
        iono_options=iono_options,
    )



from __future__ import annotations

import numpy as np
import pytest
from astropy.utils import iers

iers.conf.auto_download = False

from astropy.time import Time
from spinifex.times import (
    get_consecutive_days,
    get_gps_week,
    get_indexlist_unique_days,
    get_unique_days,
    get_unique_days_index,
)


@pytest.fixture
def times() -> Time:
    times_str = [
        "1999-02-01T00:00:00.00",
        "2010-02-01T00:00:00",
        "2010-02-01T14:27:01",
        "2024-02-01T00:00:00",
        "2024-02-01T01:00:00",
    ]
    return Time(times_str)


def test_unique_days(times) -> None:
    unique_days = get_unique_days(times)
    assert len(unique_days) == 3


def test_gps_week(times):
    gps_weeks = get_gps_week(times)
    test_weeks = np.array([995, 1569, 1569, 2299, 2299])
    assert np.all(gps_weeks == test_weeks)


def test_unique_days_index(times):
    unique_days_index = get_unique_days_index(times)
    test_index = np.array([0, 1, 1, 2, 2])
    assert np.all(unique_days_index == test_index)


def test_get_indexlist_unique_days(times):
    unique_days = get_unique_days(times)
    test_index = np.zeros((3, 5), dtype=bool)
    test_index[0, 0] = True
    test_index[1, (1, 2)] = True
    test_index[2, 3:] = True
    indexlist_unique_days = get_indexlist_unique_days(unique_days, times)
    assert np.all(indexlist_unique_days == test_index)


def test_consecutive_days(times):
    unique_days = get_unique_days(times)
    consecutive_days_index = get_consecutive_days(unique_days)
    test_index = np.array([0, 1, 2])
    assert np.all(consecutive_days_index == test_index)



from __future__ import annotations

import os
import runpy
import shutil
from pathlib import Path

import nbconvert
import nbformat
import pytest

# Make Jupyter happy
os.environ["JUPYTER_PLATFORM_DIRS"] = "1"

# Define the directory containing your example scripts
EXAMPLES_DIR = Path("docs/source/examples").resolve()


# Discover all Python files in the examples directory
example_scripts = list(EXAMPLES_DIR.glob("*.py"))

# Discover all Jupyter notebooks in the examples directory
example_notebooks = list(EXAMPLES_DIR.glob("*.ipynb"))


@pytest.mark.filterwarnings("ignore:'datfix' made the change")
@pytest.mark.parametrize("notebook", example_notebooks, ids=lambda nb: nb.name)
def test_example_notebook(notebook: Path, tmpdir):
    """Run Jupyter notebook and ensure it executes without errors."""

    # Convert notebook to a temporary Python script
    tmp_script_path = tmpdir / notebook.with_suffix(".py").name
    exporter = nbconvert.ScriptExporter()
    with notebook.open("r", encoding="utf-8") as f:
        notebook_node = nbformat.read(f, as_version=4)
    script_content, _ = exporter.from_notebook_node(notebook_node)

    with tmp_script_path.open("w", encoding="utf-8") as f:
        f.write(script_content)

    runpy.run_path(str(tmp_script_path))


@pytest.mark.filterwarnings("ignore:'datfix' made the change")
@pytest.mark.parametrize("script", example_scripts, ids=lambda s: s.name)
def test_example_script(script, tmpdir):
    """Run example script using runpy and ensure it executes without errors."""
    # copy script to temporary directory for write access
    tmp_script_path = tmpdir / script.name
    shutil.copy(script, tmp_script_path)
    runpy.run_path(str(tmp_script_path))



"""Testing of magnetic"""

from __future__ import annotations

from astropy.utils import iers

iers.conf.auto_download = False

import astropy.units as u
import numpy as np
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from spinifex.geometry.get_ipp import get_ipp_from_skycoord
from spinifex.magnetic import magnetic_models


def is_convertible_to_unit(quantity: u.Quantity, unit: u.Unit) -> bool:
    """Test if unit is convertible to a given unit"""
    try:
        _ = quantity.to(unit)
        return True
    except u.UnitsError:
        return False


def test_get_magnetic_field():
    """Test that get_magnetic does not crash"""
    cas_a = SkyCoord(ra=350.85 * u.deg, dec=58.815 * u.deg)
    lon = 6.367 * u.deg
    lat = 52.833 * u.deg
    heights = np.arange(100, 2000, 100) * u.km
    dwingeloo = EarthLocation(lon=lon, lat=lat, height=0 * u.km)
    times = Time("2020-01-20T01:00:00") + np.arange(0, 10) * 15 * u.min
    ipp = get_ipp_from_skycoord(
        loc=dwingeloo, times=times, source=cas_a, height_array=heights
    )
    field = magnetic_models.ppigrf(ipp)
    assert field.shape == times.shape + heights.shape
    assert is_convertible_to_unit(field, u.tesla)
    assert np.isclose(field[0, 0].value, -5335, 0.5)



"""Module for getting the Earth magnetic field"""

from __future__ import annotations

from spinifex.magnetic.models import MagneticFieldFunction, magnetic_models

__all__ = ["MagneticFieldFunction", "magnetic_models"]



"""Predict / correct ionospheric RM in FITS images"""

# Follows report by Van Eck (2021) and
# Python implementation in FRion https://github.com/CIRADA-Tools/FRion/
# FRion License: MIT Copyright (c) 2021 Cameron Van Eck

from __future__ import annotations

from pathlib import Path
from typing import Any, NamedTuple

import astropy.units as u
import numpy as np
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.io import fits
from astropy.time import Time, TimeDelta
from astropy.wcs import WCS
from numpy.typing import NDArray

from spinifex.exceptions import FITSHeaderError
from spinifex.get_rm import DEFAULT_IONO_HEIGHT, RM, get_rm_from_skycoord
from spinifex.image_tools.image_tools import IntegratedRM, get_integrated_rm
from spinifex.logger import logger


class FITSMetaData(NamedTuple):
    """Metadata from a FITS file"""

    start_time: Time
    """Start time of obsrvation"""
    duration: TimeDelta
    """Duration of observation"""
    location: EarthLocation
    """Location of observation"""
    name: str
    """Name of FITS file as str"""
    source: SkyCoord
    """Target of observation"""


def get_metadata_from_fits(fits_path: Path) -> FITSMetaData:
    """Get metadata from a FITS file header

    Parameters
    ----------
    fits_path : Path
        FITS file

    Returns
    -------
    FitsMetaData
        FITS image metadata
    """

    header = fits.getheader(fits_path)
    wcs = WCS(header)
    wcs_celestial = wcs.celestial
    source = SkyCoord(wcs_celestial.pixel_to_world(*wcs_celestial.wcs.crpix))

    # Lots of try/excepts/if/elses here - I'm sorry
    # This is a bit gross, but I'm being super paranoid

    if "MJD-OBS" in header:
        start_time = Time(header["MJD-OBS"], format="mjd")

    else:
        try:
            start_time = Time(
                header["DATE-OBS"],
                format="fits",
                scale="utc",
            )
        except KeyError as key_error:
            msg = "Keys `DATE-OBS` or `MJD-OBS` not in the header - we can't get the time of the observation"
            raise FITSHeaderError(msg) from key_error
    try:
        duration = TimeDelta(header["DURATION"], format="sec")
    except KeyError as key_error:
        msg = "`DURATION` not in the header - we can't get the time of the observation"
        raise FITSHeaderError(msg) from key_error

    # Get telescope location
    if "TELESCOP" in header:
        msg = "Getting telescope location from `TELESCOP` key"
        logger.info(msg)
        location = EarthLocation.of_site(site_name=header["TELESCOP"])

    elif all(key in header for key in ("OBSGEO-X", "OBSGEO-Y", "OBSGEO-Z")):
        msg = "Getting telescope location from `OBSGEO` keys"
        logger.info(msg)
        location = EarthLocation(
            x=header["OBSGEO-X"], y=header["OBSGEO-Y"], z=header["OBSGEO-Z"]
        )
    elif all(key in header for key in ("ALT-OBS", "LAT-OBS", "LONG-OBS")):
        msg = "Getting telescope location from `ALT-,LAT-,LONG-OBS` keys"
        logger.info(msg)
        location = EarthLocation(
            lon=header["LONG-LOBS"], lat=header["LAT-OBS"], height=header["ALT-OBS"]
        )
    else:
        msg = "Cannot determine telescope coordinates from FITS header"
        raise FITSHeaderError(msg)

    return FITSMetaData(
        start_time=start_time,
        duration=duration,
        location=location,
        name=fits_path.name,
        source=source,
    )


def get_rm_from_fits(
    fits_path: Path,
    timestep: u.Quantity = 15 * u.min,
    height_array: NDArray[np.float64] = DEFAULT_IONO_HEIGHT,
    iono_model_name: str = "ionex",
    magnetic_model_name: str = "ppigrf",
    **iono_kwargs: Any,
) -> RM:
    """Get the ionospheric RM from a FITS file

    Parameters
    ----------
    fits_path : Path
        Path to FITS file
    timestep : u.Quantity, optional
        Time step to evaluate time-dependant RM, by default 15*u.min
    height_array : NDArray[np.float64], optional
        Height of ionosphere, by default DEFAULT_IONO_HEIGHT
    iono_model_name : str, optional
        ionospheric model name, by default "ionex". Must be a supported ionospheric model.
    magnetic_model_name : str, optional
        geomagnetic model name, by default "ppigrf". Must be a supported geomagnetic model.
    iono_kwargs : dict
        Keyword arguments for the ionospheric model

    Returns
    -------
    RM
        Rotation measure object
    """
    fits_metadata = get_metadata_from_fits(fits_path=fits_path)

    start_time = fits_metadata.start_time
    duration = fits_metadata.duration
    end_time = start_time + duration
    times_mjd = np.arange(
        start=start_time.mjd, stop=end_time.mjd, step=timestep.to(u.day).value
    )
    times = Time(times_mjd, format="mjd")

    return get_rm_from_skycoord(
        loc=fits_metadata.location,
        times=times,
        source=fits_metadata.source,
        height_array=height_array,
        iono_model_name=iono_model_name,
        magnetic_model_name=magnetic_model_name,
        **iono_kwargs,
    )


def get_freq_from_fits(fits_path: Path) -> u.Quantity:
    """Get frequency array from FITS file

    Parameters
    ----------
    fits_path : Path
        Path to FITS file

    Returns
    -------
    u.Quantity
        Frequency array

    Raises
    ------
    FITSHeaderError
        If no spectral axis is found
    """
    header = fits.getheader(fits_path)

    if "SPECSYS" not in header:
        msg = "Could not find `SPECSYS` in header, assuming `TOPOCENT`"
        logger.warning(msg)
        header["SPECSYS"] = "TOPOCENT"

    wcs = WCS(header)
    if not wcs.has_spectral:
        msg = "FITS header has no spectral axis"
        raise FITSHeaderError(msg)

    specral_wcs = wcs.spectral

    return specral_wcs.pixel_to_world(np.arange(*specral_wcs.array_shape))


def get_integrated_rm_from_fits(
    fits_path: Path,
    timestep: u.Quantity = 15 * u.min,
    height_array: NDArray[np.float64] = DEFAULT_IONO_HEIGHT,
    iono_model_name: str = "ionex",
    magnetic_model_name: str = "ppigrf",
    **iono_kwargs: Any,
) -> IntegratedRM:
    """Computed the integrated RM effect following Van Eck (2021)

    Parameters
    ----------
    fits_path : Path
        Path to FITS file
    timestep : u.Quantity, optional
        Timestep to use for computing time-dependent RM, by default 15*u.min
    height_array : NDArray[np.float64], optional
        altitudes, by default DEFAULT_IONO_HEIGHT
    iono_model_name : str, optional
        ionospheric model name, by default "ionex". Must be a supported ionospheric model.
    magnetic_model_name : str, optional
        geomagnetic model name, by default "ppigrf". Must be a supported geomagnetic model.
    iono_kwargs : dict
        keyword arguments for the ionospheric model

    Returns
    -------
    IntegratedRM
        Integrated rotation measure object
    """

    # Get the time-dependent RM
    time_dep_rm = get_rm_from_fits(
        fits_path=fits_path,
        timestep=timestep,
        height_array=height_array,
        iono_model_name=iono_model_name,
        magnetic_model_name=magnetic_model_name,
        **iono_kwargs,
    )

    freq_arr = get_freq_from_fits(fits_path)

    return get_integrated_rm(
        time_dep_rm=time_dep_rm,
        freq_arr=freq_arr,
    )


def correct_fits_images(
    stokes_q_path: Path,
    stokes_u_path: Path,
    integrated_rm: IntegratedRM,
    ext: int = 0,
    suffix: str = "fr_corr",
) -> None:
    msg = "TODO: Implement correct_fits_images"
    raise NotImplementedError(msg)



"""Async utility functions for Spinifex."""

from __future__ import annotations

import asyncio
import sys
from collections.abc import Coroutine
from functools import wraps
from typing import Callable, TypeVar

import nest_asyncio

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


def sync_wrapper(coro: Callable[P, Coroutine[None, None, T]]) -> Callable[P, T]:
    # Apply the `nest_asyncio` magic to allow nested event loops
    # Need for running async functions in a Jupyter notebook
    nest_asyncio.apply()

    @wraps(coro)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            # Check if we are in an event loop
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # Nope, we are not in an event loop
            # Create a new event loop with `asyncio.run`
            return asyncio.run(coro(*args, **kwargs))
        # We are in an event loop
        # Use the current event loop to run the coroutine
        # This is useful when we are in a Jupyter notebook
        return loop.run_until_complete(coro(*args, **kwargs))

    return wrapper
